/* Copyright (C) 2000-2004 by George Williams */
/*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <gmodule.h>

const struct unicode_nameannot {const char *name, *annot; } * const *const *_UnicodeNameAnnot = NULL;

/*#define PROMPT_ICONV*/

#ifdef PROMPT_ICONV
# include <sys/types.h>
# include <sys/stat.h>
# include <sys/wait.h>
# include <fcntl.h>
# include <unistd.h>
#endif

typedef struct umap {
    int sel_enc;
    char utf8[12];		/* utf8 string of the selected character */
    GtkWidget *window;

    GtkWidget *close, *show, *prev, *next;
    GtkWidget *hex, *dec, *unicode;
    GtkWidget *text, *curchar;
    GtkWidget *charsets;
    GtkWidget *pagelab;
    GtkTooltips *tips;

    GtkWidget *pane;
    PangoContext *pc;
    GdkGC *gc;
    GIConv ic, ic_back;
    gchar **encnames;
    const gchar *curname;
    unsigned int is_1byte: 1;
    unsigned int is_2byte: 1;	/* Some encodings (SJIS) are both one and two bytes */
    unsigned int is_ucs2:  1;	/* Need to invert things on little endian machines */
    char *iso_2022_escape;	/* Used to shift into two byte mode */
    int   iso_2022_escape_len;
    int   low_page, high_page;	/* for 2 byte encs */
    int   cur_page;
    int   last_uni_tip;
} UMap;

GdkAtom utf8_target;

static int uni_to_enc(UMap *umap,gunichar uni) {
    gchar *utf8;
    gchar *ip, *op;
    guint il, ol;
    char outbuf[30];

    if ( uni==-1 )
return( -1 );
    if ( umap->ic_back==NULL )
return( -1 );
    utf8 = g_ucs4_to_utf8(&uni,1,NULL,NULL,NULL);
    if ( utf8==NULL )
return( -1 );
    ip = utf8;
    il = strlen(utf8);
    op = outbuf;
    ol = sizeof(outbuf);
    g_iconv(umap->ic_back,&ip,&il,&op,&ol);
    *op = 0;
    g_free(utf8);

    op = outbuf;
    if ( umap->iso_2022_escape!=NULL &&
	    strncmp(outbuf,umap->iso_2022_escape,umap->iso_2022_escape_len)==0 )
	op += umap->iso_2022_escape_len;
    if ( strlen( op )==1 )
return( (guchar) op[0] );
    else if ( strlen( op )!=2 )
return( -1 );
#if G_BYTE_ORDER==G_LITTLE_ENDIAN
    else if ( umap->is_ucs2 )
return( (((guchar *) op)[1]<<8) | ((guchar *) op)[0] );
#endif
    else
return( (((guchar *) op)[0]<<8) | ((guchar *) op)[1] );
}

static char *enc_to_utf8( UMap *umap, int slot, gchar *utf8, guint utf8len ) {
    utf8[0]='\0';
    if ( umap->ic!=NULL ) {
	char inbuf[20], *ip = inbuf, *op = utf8;
	guint il;
	if ( !umap->is_2byte || (slot<=0xff && umap->is_1byte )) {
	    inbuf[0] = slot;
	    inbuf[1] = '\0';
	    il = 1;
#if G_BYTE_ORDER==G_LITTLE_ENDIAN
	} else if ( umap->is_ucs2 ) {
	    inbuf[0] = slot&0xff;
	    inbuf[1] = slot>>8;
	    il = 2;
#endif
	} else {
	    if ( umap->iso_2022_escape!=NULL )
		strcpy(inbuf,umap->iso_2022_escape);
	    inbuf[umap->iso_2022_escape_len] = slot>>8;
	    inbuf[umap->iso_2022_escape_len+1] = slot&0xff;
	    inbuf[umap->iso_2022_escape_len+2] = 0;
	    il = umap->iso_2022_escape_len+2;
	}
	g_iconv(umap->ic,&ip,&il,&op,&utf8len);
	*op = 0;
    }
return( utf8 );
}

static gunichar enc_to_uni( UMap *umap, int slot ) {
    gchar utf8[12];

    enc_to_utf8(umap,slot,utf8,sizeof(utf8));
    if ( utf8[0]=='\0' )
return( -1 );
return( g_utf8_get_char(utf8));
}

static int TryEscape( UMap *umap,GIConv ic,char *escape_sequence ) {
    char from[20], ucs2[20];
    guint fromlen, tolen;
    char *fpt, *upt;
    int i, j, low;
    int esc_len = strlen(escape_sequence);

    strcpy(from,escape_sequence);

    low = -1;
    for ( i=0; i<256; ++i ) if ( i!=escape_sequence[0] ) {
	for ( j=0; j<256; ++j ) {
	    from[esc_len] = i; from[esc_len+1] = j; from[esc_len+2] = 0;
	    fromlen = esc_len+2;
	    fpt = from;
	    upt = ucs2;
	    tolen = sizeof(ucs2);
	    if ( g_iconv( ic , &fpt, &fromlen, &upt, &tolen )!= (size_t) (-1) &&
		    upt-ucs2==2 /* Exactly one character */ ) {
		if ( low==-1 )
		    umap->low_page = low = i;
		umap->high_page = i;
		umap->is_2byte = TRUE;
	break;
	    }
	}
    }
    if ( umap->is_2byte ) {
	umap->iso_2022_escape = escape_sequence;
	umap->iso_2022_escape_len = esc_len;
    }
return( umap->is_2byte );
}

static int GuessEscape( UMap *umap ) {
    static char escape_sequence[40];
    char utf8[20];
    guint fromlen, tolen;
    char *fpt, *tpt;
    static gunichar tests[] = { 0x3164, 0x3000, 0x2003, 0 };
    int i;

    /* The ideographic space character should be in all of the ISO-2022   */
    /*  encodings. See if we can use it to guess what the escape sequence */
    /*  is that turns on two byte mode					  */
    for ( i=0; tests[i]!=0 ; ++i ) {
	utf8[g_unichar_to_utf8(tests[i],utf8)] = '\0';
	fpt = utf8; fromlen = strlen(utf8);
	tpt = escape_sequence; tolen = sizeof(escape_sequence);
	if ( g_iconv( umap->ic_back , &fpt, &fromlen, &tpt, &tolen )!= (size_t) (-1) &&
		strlen(escape_sequence)>2 ) {
	    /* assume a two byte encoding after the escape sequence */
	    /* remove the encoding for the space, leaving just the escape */
	    /* see if that works */
	    escape_sequence[strlen(escape_sequence)-2] = '\0';
return( TryEscape( umap, umap->ic, escape_sequence ));
	}
    }

return( FALSE );
}

static void figure_encoding(UMap *umap) {
    GIConv ic =  g_iconv_open( "UCS-2", umap->curname );
    guchar good[256];
    int i, j, any, all;
    char from[8], ucs2[20];
    guint fromlen, tolen;
    char *fpt, *upt;

    umap->is_1byte = umap->is_2byte = FALSE;
    umap->low_page = -1; umap->high_page = -2;
    umap->iso_2022_escape = NULL;

    if ( ic == NULL )
	ic =  g_iconv_open( "UCS2", umap->curname );
    if ( ic == NULL )
return;

    memset(good,0,sizeof(good));
    any = FALSE; all = TRUE;
    for ( i=1; i<256; ++i ) {
	from[0] = i; from[1] = 0;
	fromlen = 1;
	fpt = from;
	upt = ucs2;
	tolen = sizeof(ucs2);
	if ( g_iconv( ic , &fpt, &fromlen, &upt, &tolen )!= (size_t) (-1)) {
	    good[i] = TRUE;
	    any = TRUE;
	} else
	    all = FALSE;
    }
    if ( any )
	umap->is_1byte = TRUE;

    if ( !all ) {
	for ( i=umap->is_1byte; i<256; ++i ) if ( !good[i] ) {
	    for ( j=0; j<256; ++j ) {
		from[0] = i; from[1] = j; from[2] = 0;
		fromlen = 2;
		fpt = from;
		upt = ucs2;
		tolen = sizeof(ucs2);
		if ( g_iconv( ic , &fpt, &fromlen, &upt, &tolen )!= (size_t) (-1) &&
			upt-ucs2==2 /* Exactly one character */ ) {
		    if ( umap->low_page==-1 )
			umap->low_page = i;
		    umap->high_page = i;
		    umap->is_2byte = TRUE;
	    break;
		}
	    }
	}
	if ( umap->low_page==umap->high_page ) {
	    umap->is_2byte = FALSE;
	    umap->low_page = umap->high_page = -1;
	}
	if ( !umap->is_2byte && !good[033]/* escape */ ) {
	    if ( TryEscape( umap,ic,"\33&@\33$B" ) ||
		    TryEscape( umap,ic,"\33$B" ) ||
		    TryEscape( umap,ic,"\33$@" ) ||
		    TryEscape( umap,ic,"\33(D" ) ||
		    TryEscape( umap,ic,"\33$(C" ) ||
		    TryEscape( umap,ic,"\33$A" ) ||
		    GuessEscape( umap ) )
		;
	}
    }
    g_iconv_close(ic);
}

static void redraw(UMap *umap) {
    GdkRectangle rect;

    rect.x = rect.y = 0;
    rect.width = rect.height = 2048;		/* A big number */
    if ( umap->pane!=NULL )
	gdk_window_invalidate_rect( GTK_WIDGET(umap->pane)->window, &rect, FALSE );
}

static void set_nextprev(UMap *umap) {
    int next_sen = FALSE, prev_sen = FALSE;
    char buffer[20];

    if ( umap->is_2byte ) {
	if ( umap->cur_page < umap->high_page )
	    next_sen = TRUE;
	if ( umap->is_1byte )
	    prev_sen = umap->cur_page>0;
	else if ( umap->cur_page> umap->low_page )
	    prev_sen = TRUE;
	sprintf( buffer, " Page: 0x%02X", umap->cur_page );
	gtk_label_set_text(GTK_LABEL(umap->pagelab), buffer);
    }
    gtk_widget_set_sensitive(umap->next,next_sen);
    gtk_widget_set_sensitive(umap->prev,prev_sen);
}

static int curchar_to_enc(UMap *umap) {
    G_CONST_RETURN gchar *val = gtk_entry_get_text( GTK_ENTRY(umap->curchar) );
    int enc;

    if ( *val=='\0' )
return( -1 );

    if (( *val=='U' || *val=='u' ) && val[1]=='+' ) {
	gunichar uni = strtol(val+2,NULL,16);
	enc = uni_to_enc(umap,uni);
    } else if ( g_utf8_strlen(val,-1)==1 && !isdigit(*val) ) {
	gunichar uni = g_utf8_get_char(val);
	enc = uni_to_enc(umap,uni);
    } else {
	enc = (val[1]=='x' || val[1]=='X' ) ? strtol(val,NULL,16) : strtol(val,NULL,10);
    }
    if ( !umap->is_2byte && enc>0xff )
return( -1 );
    else if ( enc>0xffff )
return( -1 );

return( enc );
}

static void exit_callback( GtkWidget *widget, gpointer   data ) {
    gtk_main_quit ();
}

static void show_callback( GtkWidget *widget, gpointer   data ) {
    UMap *umap = (UMap *) data;
    int enc = curchar_to_enc(umap);

    if ( enc==-1 )
return;
    umap->sel_enc = enc;
    umap->cur_page = enc>>8;
    set_nextprev(umap);
    redraw(umap);
}

static void enableshow_callback( GtkWidget *widget, gpointer   data ) {
    UMap *umap = (UMap *) data;
    int enc = curchar_to_enc(umap);

    gtk_widget_set_sensitive( umap->show , enc!=-1 );
}

static void nextprev_callback( GtkWidget *widget, gpointer data ) {
    UMap *umap = (UMap *) data;

    if ( widget==umap->next ) {
	if ( !umap->is_2byte )
	    gdk_beep();
	else if ( umap->is_1byte && umap->cur_page==0 && umap->low_page!=0 )
	    umap->cur_page = umap->low_page;
	else if ( umap->cur_page < umap->high_page )
	    ++ umap->cur_page;
	else
	    gdk_beep();
    } else {
	if ( !umap->is_2byte )
	    gdk_beep();
	else if ( umap->is_1byte && umap->cur_page!=0 && umap->low_page==umap->cur_page )
	    umap->cur_page = 0;
	else if ( umap->cur_page > umap->low_page )
	    -- umap->cur_page;
	else
	    gdk_beep();
    }

    set_nextprev(umap);
    redraw(umap);
}

static void format_current_char(UMap *umap, int enc) {
    char buffer[20];

    if ( enc==-1 )
	buffer[0] = '\0';
    else if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(umap->dec) ) )
	sprintf( buffer, "%d", enc );
    else if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(umap->hex) )) {
	if ( umap->is_1byte && enc<0x100 )
	    sprintf( buffer, "0x%02x", enc );
	else
	    sprintf( buffer, "0x%04x", enc );
    } else {
	sprintf( buffer, "U+%04x", enc_to_uni(umap,enc));
    }
    gtk_entry_set_text( GTK_ENTRY(umap->curchar),buffer );
}

static void display_callback( GtkWidget *widget, gpointer   data ) {
    UMap *umap = (UMap *) data;
    int enc = curchar_to_enc( umap );

    if ( enc==-1 )
return;

    format_current_char(umap,enc);
}

static void charset_callback( GtkWidget *widget, gpointer   data ) {
    UMap *umap = (UMap *) data;

    if ( umap->ic )
	g_iconv_close( umap->ic );
    if ( umap->ic_back )
	g_iconv_close( umap->ic_back );

#if 1
    umap->curname = gtk_entry_get_text( GTK_ENTRY( widget ));
#else
    index = gtk_option_menu_get_history( GTK_OPTION_MENU(widget));
    umap->curname = umap->encnames[index];
#endif
    umap->is_ucs2 = strcmp(umap->curname,"UCS2")==0 || strcmp(umap->curname,"UCS-2")==0;
    umap->ic = g_iconv_open( "UTF-8", umap->curname );
    umap->ic_back = g_iconv_open( umap->curname, "UTF-8" );
    figure_encoding(umap);
    if ( umap->is_1byte )
	umap->cur_page = 0;
    else
	umap->cur_page = umap->low_page;
    if ( umap->is_2byte )
	gtk_widget_show( umap->pagelab );
    else
	gtk_widget_hide( umap->pagelab );
    set_nextprev(umap);
    enableshow_callback( umap->curchar, (gpointer) umap );
    redraw(umap);
}

static void draw_uni(GdkDrawable *drw, UMap *umap, int slot,
	int x, int y ) {
    PangoLayout *layout;
    gchar utf8[12];
    gunichar uni;

    enc_to_utf8( umap, slot, utf8, sizeof(utf8) );

    if ( g_utf8_strlen(utf8,-1)!=1 )
return;

    uni = g_utf8_get_char(utf8);
    if ( (uni&0xffff)==0xffff || uni==0xfffe )	/* Not valid unicode chars */
return;						/* pango_layout complains */
    layout = gtk_widget_create_pango_layout( GTK_WIDGET(umap->pane), utf8 );
    /* How do I center the glyph? */
    gdk_draw_layout( drw, umap->gc, x+2, y, layout );
    g_object_unref( G_OBJECT(layout) );
}

static void umap_post_sel(GtkWidget *pane, GtkSelectionData *seldat,
	guint info, guint time ) {
    UMap *umap = (UMap *) g_object_get_data( G_OBJECT(pane), "umap" );
    char latin1[20];
    gunichar uni;

    switch ( info ) {
      case 0: /* utf8 data */
        gtk_selection_data_set(seldat,utf8_target,8,(gpointer) umap->utf8,strlen(umap->utf8));
      break;
      case 1: /* latin1 data */
	uni = g_utf8_get_char( umap->utf8 );
	if ( uni<=0xff ) {
	    latin1[0] = uni;
	    latin1[1] = 0;
	} else
	    sprintf( latin1, "\\{x%04x}", uni );
        gtk_selection_data_set(seldat,GDK_SELECTION_TYPE_STRING,8,(gpointer) latin1,strlen(latin1));
      break;
    }
}

static gint umap_click(GtkWidget *pane, GdkEventButton *event) {
    int i, j;
    int boxw, boxh, offx, offy;
    UMap *umap = (UMap *) g_object_get_data( G_OBJECT(pane), "umap" );

    boxw = (pane->allocation.width-1)/16;
    boxh = (pane->allocation.height-1)/16;
    offx = (pane->allocation.width - 16*boxw-1)/2;
    offy = (pane->allocation.height - 16*boxh-1)/2;

    if ( event->x<offx || event->x > offx+16*boxw ||
	    event->y < offy || event->y > offy+16*boxh )
return( TRUE );

    i = (event->y-offy)/boxh;
    j = (event->x-offx)/boxw;
    umap->sel_enc = umap->cur_page*256 + i*16 + j;
    format_current_char( umap, umap->sel_enc );

    enc_to_utf8(umap, umap->sel_enc, umap->utf8, sizeof(umap->utf8));
    if ( event->type==GDK_2BUTTON_PRESS )
	gtk_entry_set_text( GTK_ENTRY(umap->text), umap->utf8 );
    else
	gtk_entry_append_text( GTK_ENTRY(umap->text), umap->utf8 );
    redraw(umap);

    gtk_selection_owner_set(umap->pane,GDK_SELECTION_PRIMARY,event->time);
    gtk_selection_owner_set(umap->pane,GDK_SELECTION_CLIPBOARD,event->time);
return( TRUE );
}

static void utf_annot_strcpy(char *to, const char *from, int len) {
    register gunichar ch;

    while ( (ch = *(unsigned char *) from++) != '\0' && --len>=0 ) {
	if ( from[-2]=='\t' ) {
	    if ( ch=='*' ) ch = 0x2022;
	    else if ( ch=='x' ) ch = 0x2192;
	    else if ( ch==':' ) ch = 0x224d;
	    else if ( ch=='#' ) ch = 0x2245;
	    if ( len>=6 ) {
		int nl = g_unichar_to_utf8(ch,to);
		len -= nl;
		to += nl;
	    }
	} else
	    *(to++) = ch;
    }
    *to = 0;
}

static gint umap_motion(GtkWidget *pane, GdkEventMotion *event) {
    int i, j;
    int boxw, boxh, offx, offy;
    UMap *umap = (UMap *) g_object_get_data( G_OBJECT(pane), "umap" );
    gunichar uni;
    char space[300];

    boxw = (pane->allocation.width-1)/16;
    boxh = (pane->allocation.height-1)/16;
    offx = (pane->allocation.width - 16*boxw-1)/2;
    offy = (pane->allocation.height - 16*boxh-1)/2;

    if ( event->x<offx || event->x > offx+16*boxw ||
	    event->y < offy || event->y > offy+16*boxh ) {
	gtk_tooltips_set_tip( GTK_TOOLTIPS(umap->tips), umap->pane, "", NULL );
	umap->last_uni_tip = -1;
return( TRUE );
    }

    i = (event->y-offy)/boxh;
    j = (event->x-offx)/boxw;
    uni = enc_to_uni( umap, umap->cur_page*256 + i*16 + j );
    if ( umap->last_uni_tip == uni )
	/* Do Nothing */;
    else if ( uni==-1 || uni>= 0x110000 ) {
	gtk_tooltips_set_tip( GTK_TOOLTIPS(umap->tips), umap->pane, "", NULL );
	umap->last_uni_tip = -1;
    } else {
	sprintf( space, " U+%04X ", uni );
	if ( _UnicodeNameAnnot!=NULL &&
		_UnicodeNameAnnot[uni>>16][(uni>>8)&0xff][uni&0xff].name!=NULL ) {
	    strcpy(space+strlen(space),
		    _UnicodeNameAnnot[uni>>16][(uni>>8)&0xff][uni&0xff].name);
	}
	if ( _UnicodeNameAnnot!=NULL &&
		_UnicodeNameAnnot[uni>>16][(uni>>8)&0xff][uni&0xff].annot!=NULL ) {
	    strcat(space,"\n");
	    utf_annot_strcpy(space+strlen(space),
		    _UnicodeNameAnnot[uni>>16][(uni>>8)&0xff][uni&0xff].annot,
		    sizeof(space)-strlen(space));
	}
	gtk_tooltips_set_tip( GTK_TOOLTIPS(umap->tips), umap->pane, space, NULL );
	umap->last_uni_tip = uni;
    }
return( TRUE );
}

static gint umap_expose (GtkWidget *pane, GdkEventExpose *event) {
    int i, j;
    int boxw, boxh, offx, offy;
    int as = 0;
    GdkDrawable *drw = GDK_DRAWABLE(pane->window);
    UMap *umap = (UMap *) g_object_get_data( G_OBJECT(pane), "umap" );

    if ( umap->gc==NULL ) {
	umap->gc = gdk_gc_new( drw );
	umap->pc = gtk_widget_get_pango_context( GTK_WIDGET(umap->pane) );
    }

    boxw = (pane->allocation.width-1)/16;
    boxh = (pane->allocation.height-1)/16;
    offx = (pane->allocation.width - 16*boxw-1)/2;
    offy = (pane->allocation.height - 16*boxh-1)/2;
    as = 0;					/* for a layout */

    if ( umap->cur_page==(umap->sel_enc>>8) ) {
	GtkStyle *style = gtk_widget_get_style(umap->text);
	GdkGCValues values;
	i = (umap->sel_enc&0xf0)>>4;
	j = (umap->sel_enc&0x0f);
	gdk_gc_get_values(umap->gc,&values);
	gdk_gc_set_foreground(umap->gc,&style->fg[GTK_STATE_SELECTED]);
	gdk_draw_rectangle(drw,umap->gc,TRUE,
		offx+j*boxw, offy+i*boxh, boxw, boxh );
	gdk_gc_set_foreground(umap->gc,&values.foreground);
    }

    gdk_draw_rectangle(drw,umap->gc,FALSE,offx,offy, 16*boxw+1, 16*boxh+1);

    for ( i=1; i<16; ++i ) {
	gdk_draw_line(drw,umap->gc,
		offx+i*boxw, offy,
		offx+i*boxw, offy+16*boxh );
	gdk_draw_line(drw,umap->gc,
		offx, offy+i*boxh,
		offx+16*boxw, offy+i*boxh );
    }

    for ( i=0; i<16; ++i ) {
	for ( j=0; j<16; ++j ) {
	    draw_uni(drw, umap, (umap->cur_page<<8) + i*16+j, offx+j*boxw, offy+i*boxh + as );
	}
    }
return( TRUE );
}

/* These are the encodings documented to exist in libiconv on the mac */
/*  I hope that means that any iconv will support them all */
static gchar *fallback_encodings[] = {
	"ASCII", 
	"ARMSCII-8",
	"BIG5",
	"BIG5-HKSCS",
	"CP850",
	"CP862",
	"CP866",
	"CP874",
	"CP932",
	"CP949",
	"CP950",
	"CP1133",
	"CP1250",
	"CP1251",
	"CP1252",
	"CP1253",
	"CP1254",
	"CP1255",
	"CP1256",
	"CP1257",
	"CP1258",
	"EUC-CN",
	"EUC-JP",
	"EUC-KR",
	"GBK",
	"Georgian-Academy",
	"Georgian-PS",
	"HP-ROMAN8",
/*	"HZ", */		/* This doesn't fit any of our catagories */
	"ISO-2022-CN",
/*	"ISO-2022-CN-EXT",	*/
	"ISO-2022-JP",
/*	"ISO-2022-JP-1",	*/
	"ISO-2022-JP-2",
	"ISO-2022-KR",
	"ISO-8859-1",
	"ISO-8859-2",
	"ISO-8859-3",
	"ISO-8859-4",
	"ISO-8859-5",
	"ISO-8859-6",
	"ISO-8859-7",
	"ISO-8859-8",
	"ISO-8859-9",
	"ISO-8859-10",
	"ISO-8859-13",
	"ISO-8859-14",
	"ISO-8859-15",
	"ISO-8859-16",
	"JOHAB",
	"KOI8-R",
	"KOI8-U",
/*	"KOI8-RU",		*/
#if 0			/* listed on mac, but not available on redhat 9 */
	"MacRoman",
	"MacCentralEurope",
	"MacIceland",
	"MacCroation",
	"MacRomania",
	"MacCyrillic",
	"MacUkraine",
	"MacGreek",
	"MacTurkish",
	"MacHebrew",
	"MacArabic",
	"MacThai",
#endif
	"Macintosh",
/*	"MuleLao-1",		*/
/*	"NEXTSTEP",		*/
	"SHIFT-JIS",
	"TCVN",
	"TIS-620",
	"UCS-2",
	"VISCII",
	NULL
};

static gchar **GetIconvEncodings(void) {
#ifdef PROMPT_ICONV
    int filedes[2];
    int pid;

    if ( pipe(filedes)==-1 )
 goto failure;
    if ( (pid = fork())== -1 ) {
	close(filedes[0]);
	close(filedes[1]);
 goto failure;
    } else if ( pid==0 ) {
	/* child */
	close(1);
	close(2);
	dup2(filedes[1],1);
	close(filedes[1]);
	close(filedes[0]);
	open("/dev/null",O_WRONLY);
	execlp("iconv","iconv", "--list", NULL );
	exit(1);			/* Should never get here */
    } else {
	char buffer[200];
	char **names = malloc(100*sizeof(char *));
	int cur=0, max=100;
	FILE *icnv = fdopen(filedes[0],"r");
	char *pt;

	close(filedes[1]);
	while ( fgets(buffer,sizeof(buffer),icnv)!=NULL ) {
	    if ( (pt = strstr(buffer,"//"))==NULL )
	continue;
	    if ( cur>=max )
		names = realloc(names,(max+=100)*sizeof(char *));
	    *pt = '\0';
	    names[cur++] = strdup(buffer);
	}
	if ( cur>=max )
	    names = realloc(names,(++max)*sizeof(char *));
	names[cur] = NULL;
	fclose(icnv);
	waitpid(pid,NULL,WNOHANG);	/* If the pipe gives EOF, then child is dead */
	if ( cur!=0 )
return( names );
	free(names);
    }
 failure:
#endif
return( fallback_encodings );
}

#if 1
static GList *CharsetList(UMap *umap) {
    GList *list = NULL;
    int i;
    gchar **names = GetIconvEncodings();

    umap->encnames = names;

    for ( i=0; names[i]!=NULL; ++i )
	list = g_list_append(list,names[i]);
return( list );
}
#else
static GtkWidget *CharsetMenu(UMap *umap) {
    GtkWidget *menu, *item;
    int i;
    gchar **names = GetIconvEncodings();

    umap->encnames = names;

    menu = gtk_menu_new ();

    for ( i=0; names[i]!=NULL; ++i ) {
	item = gtk_menu_item_new_with_label( names[i] );
	gtk_widget_show(item);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
    }
  
return( menu );
}
#endif

static void CreateCharMap(void) {
    GtkWidget *hbox1, *hbox2, *hbox3, *hboxlab;
    GtkWidget *vbox1, *vbox2, *vbox3, *vbox4, *vbox5, *buttonbox;
    GtkWidget *separator, *lab;
    UMap *umap = calloc(1,sizeof(UMap));

    umap->sel_enc      = -1;
    umap->last_uni_tip = -1;

    umap->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (umap->window), "Unicode map");
    g_signal_connect(G_OBJECT (umap->window), "delete_event",
		      G_CALLBACK (exit_callback), umap);

    buttonbox = gtk_vbox_new (FALSE, 5);
    gtk_container_add (GTK_CONTAINER (umap->window), buttonbox);
    gtk_container_set_border_width(GTK_CONTAINER (buttonbox), 5);
    gtk_widget_show (buttonbox);


    umap->text = gtk_entry_new();
    gtk_widget_show(umap->text);

    umap->close = gtk_button_new_with_mnemonic( "C_lose" );
    g_signal_connect(G_OBJECT (umap->close), "clicked",
		      G_CALLBACK (exit_callback), umap );
    gtk_widget_show(umap->close);

    hbox2 = gtk_hbox_new (FALSE, 5);
    gtk_widget_show (hbox2);
    gtk_box_pack_start (GTK_BOX (hbox2), umap->text, TRUE, TRUE, 0);
    gtk_box_pack_end (GTK_BOX (hbox2), umap->close, FALSE, FALSE, 0);

    gtk_box_pack_start (GTK_BOX (buttonbox), hbox2, FALSE, FALSE, 0);

    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (buttonbox), separator, FALSE, TRUE, 0);
    gtk_widget_show (separator);


    umap->show = gtk_button_new_with_mnemonic( "_Show" );
    g_signal_connect(G_OBJECT (umap->show), "clicked",
		      G_CALLBACK (show_callback), umap );
    gtk_widget_set_sensitive(umap->show,FALSE);
    gtk_widget_show(umap->show);

    vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox1);
    gtk_box_pack_start (GTK_BOX (vbox1), umap->show, FALSE, FALSE, 0);


    umap->hex = gtk_radio_button_new_with_mnemonic( NULL, "_Hex" );
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(umap->hex),TRUE);
    g_signal_connect(G_OBJECT (umap->hex), "toggled",
		      G_CALLBACK (display_callback), umap );
    gtk_widget_show(umap->hex);

    umap->dec = gtk_radio_button_new_with_mnemonic_from_widget(
	    GTK_RADIO_BUTTON(umap->hex), "_Dec" );
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(umap->dec),FALSE);
    g_signal_connect(G_OBJECT (umap->dec), "toggled",
		      G_CALLBACK (display_callback), umap );
    gtk_widget_show(umap->dec);

    umap->unicode = gtk_radio_button_new_with_mnemonic_from_widget(
	    GTK_RADIO_BUTTON(umap->hex), "_Unicode" );
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(umap->unicode),FALSE);
    g_signal_connect(G_OBJECT (umap->unicode), "toggled",
		      G_CALLBACK (display_callback), umap );
    gtk_widget_show(umap->unicode);

    lab = gtk_label_new_with_mnemonic("_Character:");
    /* gtk_label_set_justify(GTK_LABEL(lab),GTK_JUSTIFY_LEFT); */
    gtk_widget_show(lab);
    hboxlab = gtk_hbox_new (FALSE, 5);
    gtk_widget_show (hboxlab);
    gtk_box_pack_start (GTK_BOX (hboxlab), lab, FALSE, FALSE, 0);
    umap->curchar = gtk_entry_new();
    gtk_label_set_mnemonic_widget(GTK_LABEL(lab), umap->curchar);
    g_signal_connect(G_OBJECT (umap->curchar), "changed",
		      G_CALLBACK (enableshow_callback), umap );
    gtk_widget_show(umap->curchar);

    vbox2 = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox2);
    gtk_box_pack_start (GTK_BOX (vbox2), hboxlab, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox2), umap->curchar, FALSE, FALSE, 0);

    vbox3 = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox3);
    gtk_box_pack_start (GTK_BOX (vbox3), umap->hex, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox3), umap->dec, FALSE, FALSE, 0);

    vbox4 = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox4);
    gtk_box_pack_start (GTK_BOX (vbox4), umap->unicode, FALSE, FALSE, 0);

    hbox1 = gtk_hbox_new (FALSE, 5);
    gtk_widget_show (hbox1);
    gtk_box_pack_start (GTK_BOX (hbox1), vbox2, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox1), vbox3, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox1), vbox4, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (hbox1), vbox1, FALSE, FALSE, 0);


    /* Need to be created before option menu */
    umap->prev = gtk_button_new_with_mnemonic( "< _Prev" );
    umap->next = gtk_button_new_with_mnemonic( "_Next >" );
    umap->pagelab = gtk_label_new("");


    lab = gtk_label_new_with_mnemonic("Charact_er Set:");
    /*gtk_label_set_justify(GTK_LABEL(lab),GTK_JUSTIFY_LEFT);*/
    gtk_widget_show(lab);
    hboxlab = gtk_hbox_new (FALSE, 5);
    gtk_widget_show (hboxlab);
    gtk_box_pack_start (GTK_BOX (hboxlab), lab, FALSE, FALSE, 0);
#if 1
    umap->charsets = gtk_combo_new();
    gtk_label_set_mnemonic_widget(GTK_LABEL(lab), umap->charsets);
    g_signal_connect(G_OBJECT( GTK_COMBO(umap->charsets)->entry ), "changed",
		      G_CALLBACK (charset_callback), umap );
    gtk_combo_set_popdown_strings(GTK_COMBO(umap->charsets),CharsetList(umap));
#else
    umap->charsets = gtk_option_menu_new();
    g_signal_connect(G_OBJECT (umap->charsets), "changed",
		      G_CALLBACK (charset_callback), umap );
    gtk_option_menu_set_menu(GTK_OPTION_MENU(umap->charsets),CharsetMenu(umap));
#endif
    gtk_widget_show(umap->charsets);

    vbox5 = gtk_vbox_new (FALSE, 5);
    gtk_widget_show (vbox5);
    gtk_box_pack_start (GTK_BOX (vbox5), hboxlab, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox5), umap->charsets, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox5), hbox1, FALSE, FALSE, 0);

    gtk_box_pack_start (GTK_BOX (buttonbox), vbox5, FALSE, FALSE, 0);


    hbox3 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox3);

    g_signal_connect(G_OBJECT (umap->prev), "clicked",
		      G_CALLBACK (nextprev_callback), umap );
    gtk_widget_show(umap->prev);

    g_signal_connect(G_OBJECT (umap->next), "clicked",
		      G_CALLBACK (nextprev_callback), umap );
    gtk_widget_show(umap->next);
    gtk_box_pack_start(GTK_BOX (hbox3), umap->prev, FALSE, FALSE, 20);
    gtk_box_pack_start(GTK_BOX (hbox3), umap->pagelab, FALSE, FALSE, 20);
    gtk_box_pack_end(GTK_BOX (hbox3), umap->next, FALSE, FALSE, 20);

    gtk_box_pack_start (GTK_BOX (buttonbox), hbox3, FALSE, FALSE, 0);


    separator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (buttonbox), separator, FALSE, TRUE, 0);
    gtk_widget_show (separator);

    umap->pane = gtk_drawing_area_new ();
    gtk_widget_set_events(GTK_WIDGET(umap->pane),
	    GDK_EXPOSURE_MASK|GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK|
	    GDK_POINTER_MOTION_MASK|GDK_ENTER_NOTIFY_MASK|GDK_LEAVE_NOTIFY_MASK);
    gtk_widget_show(umap->pane);
    g_signal_connect (G_OBJECT (umap->pane), "expose_event",
		    G_CALLBACK (umap_expose), NULL);
    g_signal_connect (G_OBJECT (umap->pane), "button_press_event",
		    G_CALLBACK (umap_click), NULL);
    g_signal_connect (G_OBJECT (umap->pane), "motion_notify_event",
		    G_CALLBACK (umap_motion), NULL);
    g_signal_connect (G_OBJECT (umap->pane), "selection_get",
		    G_CALLBACK (umap_post_sel), NULL);
    gtk_selection_add_target(umap->pane,GDK_SELECTION_PRIMARY,utf8_target,0);
    gtk_selection_add_target(umap->pane,GDK_SELECTION_CLIPBOARD,utf8_target,0);
    gtk_selection_add_target(umap->pane,GDK_SELECTION_PRIMARY,GDK_SELECTION_TYPE_STRING,1);
    gtk_selection_add_target(umap->pane,GDK_SELECTION_CLIPBOARD,GDK_SELECTION_TYPE_STRING,1);
    gtk_widget_set_size_request (GTK_WIDGET (umap->pane), 17*16+2, 17*16+4 );
    gtk_box_pack_start (GTK_BOX (buttonbox), umap->pane, FALSE, FALSE, 0);
    g_object_set_data(G_OBJECT( umap->pane ), "umap", umap );


    umap->tips = gtk_tooltips_new();
    gtk_tooltips_set_tip( GTK_TOOLTIPS(umap->tips), umap->pane, "", NULL );
    /*gtk_tooltips_disable( GTK_TOOLTIPS(umap->tips) );*/


    gtk_widget_show(umap->window);
}

static void inituninameannot(void) {
#if _NO_LIBUNINAMESLIST
    _UnicodeNameAnnot = NULL;
#elif defined(_STATIC_LIBUNINAMESLIST) || defined(NODYNAMIC)
    _UnicodeNameAnnot = &UnicodeNameAnnot;
#else
    if ( g_module_supported() ) {
	GModule *libuninames=NULL;
# ifdef LIBDIR
	libuninames = g_module_open( LIBDIR "/libuninameslist.so",G_MODULE_BIND_LAZY);
# endif
	if ( libuninames==NULL )
	    libuninames = g_module_open( "libuninameslist.so",G_MODULE_BIND_LAZY);
	if ( libuninames!=NULL )
	    g_module_symbol(libuninames,"UnicodeNameAnnot",
		    (gpointer *) &_UnicodeNameAnnot);
#endif
    }
}

int main( int argc, char **argv ) {

    gtk_init (&argc, &argv);
    inituninameannot();
    gtk_rc_parse(".umap.rc");
    utf8_target = gdk_atom_intern("UTF8_STRING",FALSE);

    CreateCharMap();

    gtk_main ();
    
    return 0;
}
